from os import getenv

HARMONIC_OSCILLATOR_MAX_N = int(getenv("TRANSIT_HARMONIC_OSCILLATOR_MAX_N", 50))
SMALL_NUMBER = float(getenv("TRANSIT_SMALL_NUMBER", 1e-8))
LARGE_NUMBER = float(getenv("TRANSIT_LARGE_NUMBER", 1000))